from setuptools import setup, find_packages
import os

# 读取README.md文件作为长描述
with open(os.path.join(os.path.dirname(__file__), 'README.md'), 'r', encoding='utf-8') as f:
    long_description = f.read()

# 读取requirements.txt文件作为依赖
with open(os.path.join(os.path.dirname(__file__), 'requirements.txt'), 'r', encoding='utf-8') as f:
    requirements = [line.strip() for line in f if line.strip() and not line.startswith('#')]

setup(
    name="praasper",
    version='0.5.5',
    author="Tony Liu",
    author_email="paradeluxe3726@gmail.com",
    description="A tool for automatic speech recognition and annotation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/praasper/praasper",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=requirements,

)